# -*- coding: utf-8 -*-
"""

 MODULE:       hru-delin_basins

 AUTHOR(S):    adapted from GRASS-HRU (ILMS) - JENA University
               by IRSTEA - Christine Barachet 

 PURPOSE:    1. Relocates the gauges on the reaches
             2. Calculates watersheds at the gauges

"""

import subprocess, string, os, sys, glob, types, time
import numpy as np
import ConfigParser
import grass.script as grass
import struct
import math
import csv

from osgeo import gdal
from gdalconst import *
from osgeo import ogr


def write_log(msg, buff, start, end, best_x, best_y):
    horiz_shift = best_x - start - end 
    vertic_shift = best_y - start - end + 1
    if horiz_shift != 0 or vertic_shift != 0:
        if horiz_shift > 0:
            dir_x_shift = 'East'
        else:
            dir_x_shift = 'West'
    
        if vertic_shift > 0:
            dir_y_shift = 'South'
        else:
            dir_y_shift = 'North'
    
        logf.writerow({'gauge code' : gauge_code, 'area' : str(gauge_area), 'rule' :  msg, 
                       'horiz dir' : dir_x_shift,
                       'horiz shift' : str(abs(horiz_shift)),
                       'vertic dir' : dir_y_shift, 
                       'vertic shift' : str(abs(vertic_shift))})
                       

def snapping_gauges_to_reaches():
    """
    Relocates gauges on the reaches derivated from MNT in step 1 (r.watershed)
    """
    basin_min_size = int(parms.get('basin_min_size', 'size'))
    surface_ok_1 = int(parms.get('auto_relocation', 'surface_tolerance_1'))
    dist_ok_1 = int(parms.get('auto_relocation', 'distance_tolerance_1' ))

    if parms.get('auto_relocation', 'surface_tolerance_2') == "":
        nb_trials = 1
        surface_ok_2 = 0
        dist_ok_2 = 0
    else:
        nb_trials = 2
        surface_ok_2 = int(parms.get('auto_relocation', 'surface_tolerance_2'))
        dist_ok_2 = int(parms.get('auto_relocation', 'distance_tolerance_2' ))
    
    col_area = parms.get('auto_relocation', 'gauge_area_col_name')
    if col_area == "":
        print(' =========> ERROR ')
        sys.exit(' =========> CHECK gauge_area_col_name PARAMETER')

    gauge_area_unit = parms.get('auto_relocation', 'gauge_area_unit')
    if gauge_area_unit == '1':
        fact = 1000
        unit = 'm2'
    elif gauge_area_unit == '2':
        fact = 1000000
        unit = 'km2'
    else:
        print(' =========> ERROR ')
        sys.exit(' =========> CHECK gauge_area_unit PARAMETER')

    global logf, gauge_code, gauge_area
    logfile_n = directory_out + '/auto_reloc.csv'
    logfile = open(logfile_n, 'w')
    fieldnames = ['gauge code','area','rule','horiz shift','horiz dir','vertic shift','vertic dir']
    logf = csv.DictWriter(logfile, fieldnames=fieldnames, delimiter=';')
    logf.writeheader()

    # Read the accumulation raster and get the geo informations
    accum_file = directory_out + '/step1_accum.tif'
    accum_lyr = gdal.Open(accum_file, GA_ReadOnly)
    accum_bd = accum_lyr.GetRasterBand(1)
    georef = accum_lyr.GetGeoTransform()
    hres = georef[1]
    vres = georef[5]
    pixel_sz = (hres * abs(vres))
    acc_rast_cols = accum_lyr.RasterXSize
    acc_rast_rows = accum_lyr.RasterYSize
       
    # Matrix of euclidian distances from the center (gauge position)
    max_dist = max(dist_ok_1, dist_ok_2)
    euclid_dist = np.full((max_dist * 2 + 1, max_dist * 2 + 1), 0)
    for y in range (0, (max_dist * 2 + 1)):
                for x in range (0, (max_dist * 2 + 1)):
                   euclid_dist [y, x] = int(math.sqrt((y - max_dist)**2 + (x - max_dist)**2))
  
    # Gauges loop                                    
    for gauge in reloc_lyr:
        gauge_code = gauge.GetField(col_name)
        # Get the drained surface of the gauge
        gauge_area = gauge.GetField(col_area)
        nb_pixel_area = ((gauge_area * fact) / pixel_sz) + 0.5
        
        if nb_pixel_area > basin_min_size:
            # Get the coordinates of the gauge and calculate center of the offset
            geom = gauge.GetGeometryRef()
            gauge_x, gauge_y = geom.GetX(), geom.GetY()
            Xreste = gauge_x % hres
            Yreste = gauge_y % abs(vres)
            new_gauge_x = (gauge_x - Xreste) + 1
            new_gauge_y = (gauge_y - Yreste) - 1
          
            # Read the currents of accum raster around the gauge
            dist_max = max(dist_ok_1, dist_ok_2)
            pix_x = max(int(((new_gauge_x - georef[0]) / hres) - (dist_max)), 0)
            pix_y = max(int(((georef[3] - new_gauge_y) / abs(vres)) - (dist_max)), 0)

            #        and don't access beyond the accum raster bounds
            nb_cols = min((dist_max * 2) + 1, acc_rast_cols - pix_x)
            nb_rows = min((dist_max * 2) + 1, acc_rast_rows - pix_y)
            zone = accum_bd.ReadAsArray(pix_x, pix_y, nb_cols, nb_rows).astype(int)
            
            # Matrix of differences 
            zone_diff = abs(nb_pixel_area - abs(zone))
 
            # Prepare for first search
            accum_toler = int(nb_pixel_area * surface_ok_1 / 100)
            if dist_ok_1 > dist_ok_2:
                start = 0
                col_max = min(nb_cols, acc_rast_cols - pix_x)
                row_max = min(nb_rows, acc_rast_rows - pix_y)
                end = dist_ok_1 - start
            else:
                start = dist_ok_2 - dist_ok_1
                col_max = min(nb_cols - start, acc_rast_cols - pix_x)
                row_max = min(nb_rows - start, acc_rast_rows - pix_y)
                end = dist_ok_2 - start
            msg = 'shifted with first rule'
            buff = dist_ok_1
            trial = 1

            while trial <= nb_trials:
                best_area = zone_diff.max()
                best_euclid_dist = euclid_dist[start, start]
                best_x = 0
                best_y = 0
                better_place_found = False
               
                for y in range (start, row_max):
                    for x in range (start, col_max):
                        
                        if (zone_diff[y, x] <= accum_toler):
                            if euclid_dist[y, x] < best_euclid_dist:
                                best_area = zone_diff[y, x]
                                best_euclid_dist = euclid_dist[y, x]
                                best_y = y
                                best_x = x
                                better_place_found = True
                
                if better_place_found:
                    new_x = ((pix_x + best_x) * hres) + georef[0] + (hres / 2)
                    new_y = georef[3] - ((pix_y + best_y) * abs(vres)) - (abs(vres) / 2)
                    wkt   = "POINT(%f %f)" %  (new_x , new_y)
                    point = ogr.CreateGeometryFromWkt(wkt)
                    gauge.SetGeometryDirectly(point)
                    reloc_lyr.SetFeature(gauge)
                    trial = 3
                    write_log(msg, buff, start, end, best_x, best_y)
                else:
                    if trial == 2:
                        logf.writerow({'gauge code' : gauge_code, 'area' : str(gauge_area), 'rule' : ' no better place found'})
                        trial = 3
                    else:
                        if nb_trials == 1:
                            logf.writerow({'gauge code' : gauge_code, 'area' : str(gauge_area), 'rule' : ' no better place found'})
                            trial = 3
                        else:
                            # Prepare for second search
                            msg = 'shifted with second rule'                            
                            accum_toler = int(nb_pixel_area * surface_ok_2 / 100) 
                            trial = 2
                            buff = dist_ok_2
                            if dist_ok_2 > dist_ok_1:
                                start = 0
                                end = dist_ok_2 - start
                            else:
                                start = dist_ok_1 - dist_ok_2
                                col_max = min(nb_cols - start, acc_rast_cols - pix_x)
                                row_max = min(nb_rows - start, acc_rast_rows - pix_y )
                                end = dist_ok_1 - start
        else:
           gaugeID = gauge.GetFID()
           reloc_lyr.DeleteFeature(gaugeID)
           logf.writerow({'gauge code' : gauge_code, 'area' : str(gauge_area), 'rule' : ' basin too small ---> DELETED'})
   
    logfile.close

def cut_streams_at_gauges():
    """
    Streams generated at step 1 (by r.watershed) are cutted at the gauges
    """
    streams_file = directory_out + '/step1_streams.tif'
    streams_lyr = gdal.Open(streams_file, GA_ReadOnly)
    streams_band = streams_lyr.GetRasterBand(1)
    stream_orig = streams_band.ReadAsArray()
    
    mask_file = directory_out + '/step2_mask.tif'
    mask_lyr = gdal.Open(mask_file, GA_ReadOnly)
    mask_band = mask_lyr.GetRasterBand(1)
    mask = mask_band.ReadAsArray()
    nodata = mask_band.GetNoDataValue()
    mask[mask == nodata] = 0

    stream_orig = stream_orig * mask
 
    streams_cols = streams_lyr.RasterXSize
    streams_rows = streams_lyr.RasterYSize
    georef = streams_lyr.GetGeoTransform()
    hres = georef[1]
    vres = georef[5]

    streams_new_file = directory_out + '/step2_streams_new.tif'
    driver = gdal.GetDriverByName('GTiff')
    dst_ds = driver.Create(streams_new_file, streams_cols, streams_rows, 1, GDT_Int32)
    dst_ds.SetGeoTransform((georef[0], hres, 0, georef[3], 0, vres))
    dst_ds.SetProjection(streams_lyr.GetProjection())

    drain_file = directory_out + '/step1_drain.tif'
    drain_lyr = gdal.Open(drain_file, GA_ReadOnly)
    drain = drain_lyr.GetRasterBand(1).ReadAsArray()
    drain_dir = [[-1,1],[-1,0],[-1,-1],[0,-1],[1,-1],[1,0],[1,1],[0,1]]

    # Give possibility of more than 2 segments in a stream
    stream_wk = np.zeros((streams_rows, streams_cols), dtype=('int64'))
    nodata = streams_band.GetNoDataValue()
    for y in range (0, streams_rows - 1):
        for x in range (0, streams_cols - 1):    
            if stream_orig[y, x] != nodata:
                stream_wk[y, x] = stream_orig[y, x] * 100

    gauges_file = directory_out + '/step2_gauges_for_watersheds.shp'
    gauges_in = ogr.Open(gauges_file)
    gauges_lyr = gauges_in.GetLayer()

    for gauge in gauges_lyr:
        gauge_code = gauge.GetField(col_name)
        geom = gauge.GetGeometryRef()
        current_x = int(((geom.GetX() - georef[0]) / hres))
        current_y = int(((georef[3] - geom.GetY()) / abs(vres)))
        current_stream = stream_orig[current_y, current_x]
        
        if current_stream == nodata:
            print('---------- gauge outside of stream network : ', gauge_code)
            end_of_stream = True
        else:
            stream_wk[current_y, current_x] += 1
            end_of_stream = False
            i = 0
            d = abs(drain[current_y, current_x])

        while not end_of_stream:
            if d == 8:
                d = 0
            y = current_y + drain_dir[d][0]
            x = current_x + drain_dir[d][1]
            d2 = abs(drain[y, x]) - 1
            drain_into_y = y + drain_dir[d2][0]
            drain_into_x = x + drain_dir[d2][1]
            if (stream_orig[y, x] == current_stream) and (drain_into_y == current_y) and (drain_into_x == current_x):
                stream_wk[y, x] += 1
                current_y = y
                current_x = x
                i = 0
                d = abs(drain[y, x])
            else:
                i += 1
                if i == 7:
                    end_of_stream = True    
                else:
                    d += 1
                     
    dst_ds.GetRasterBand(1).WriteArray(stream_wk)
    dst_ds = None

'''

 MAIN


'''
try:
    from subprocess import DEVNULL # Python 3.
except ImportError:
    DEVNULL = open(os.devnull, 'wb')
    #DEVNULL = open('trace_step2', 'wb')
print('---------- HRU-delin Step 2 started ---------------------------------------------')

global logf, gauge_code, gauge_area, reloc_lyr, col_name

parms = ConfigParser.ConfigParser(allow_no_value=True)
parms_file = "hrudelin_config.cfg"
parms.read(parms_file)
directory_out = parms.get("dir_out", "files")
col_name = parms.get('for_watershed_id', 'col_name')
if col_name == "":
    print(' =========> ERROR ')
    sys.exit(' =========> CHECK col_name PARAMETER')

# Get the shape of gauges 
gauges_file = parms.get("auto_relocation", "relocated_gauges")
if gauges_file == "":
    gauges_file = directory_out + '/gauges_selected.shp'
gauges_in = ogr.Open(gauges_file)
gauges_lyr = gauges_in.GetLayer()
 
# Set the new shape
gauges_reloc_name = 'step2_gauges_for_watersheds'
gauges_reloc_file = directory_out + '/' + gauges_reloc_name + '.shp'
driver = ogr.GetDriverByName('ESRI Shapefile') 
if os.path.exists(gauges_reloc_file):
    driver.DeleteDataSource(gauges_reloc_file)
reloc_shp = driver.CreateDataSource(gauges_reloc_file)
reloc_lyr = reloc_shp.CopyLayer(gauges_lyr, gauges_reloc_name)

# Relocation of the gauges
if (parms.get("auto_relocation", "to_do")) == 'yes':
    print('---------- HRU-delin Step 2 : Relocation of the gauges')
    snapping_gauges_to_reaches()
reloc_shp.ExecuteSQL('REPACK ' + reloc_lyr.GetName())
reloc_shp.Destroy()

# Import drain raster
drain_layer = directory_out + '/step1_drain.tif'
drain_wk = 'drain_wk'
grass.run_command('r.in.gdal', flags='o', input=drain_layer, output=drain_wk, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('g.proj', flags='p', georef='drain_wk', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('g.region', flags='sp', rast=drain_wk, stdout = DEVNULL, stderr = DEVNULL)

# Watersheds derivation 
basins='basins'
rasters_list='basins'
# r.cross doesn't accept more than 30 layers
max_rasters=29
i=0
grass.run_command('r.mapcalc', basins='null()', stdout = DEVNULL, stderr = DEVNULL)
ds = ogr.Open(gauges_reloc_file)
reloc_lyr = ds.GetLayer()
nb_gauges = reloc_lyr.GetFeatureCount()

print('---------- HRU-delin Step 2 : Derivating watersheds at gauges')
for gauge in reloc_lyr:
    geom = gauge.GetGeometryRef()
    gauge_x, gauge_y = geom.GetX(), geom.GetY()
    basin_wk = 'basin_wk'
    grass.run_command('r.water.outlet', drainage=drain_wk, basin='basin_tmp%d' %(i),
                       easting = gauge_x, northing = gauge_y, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
    rasters_list += ',basin_tmp%d' %(i)
    i += 1
    if i % max_rasters == 0 or i == nb_gauges:
        grass.run_command('r.cross', input=rasters_list, output=basins, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
        grass.run_command('g.remove', flags='f', rast='basin_tmp*', stdout = DEVNULL, stderr = DEVNULL)
        rasters_list='basins'  
  
grass.run_command('r.null', map=basins, setnull=0, stdout = DEVNULL, stderr = DEVNULL)

print('---------- HRU-delin Step 2 : Reclassifying the watersheds')
recl_watersheds_file = 'reclass_watersheds_rules'
rules_ws = open(recl_watersheds_file, "w")
ds = ogr.Open(gauges_reloc_file)
reloc_lyr = ds.GetLayer()

for gauge in reloc_lyr:
    geom = gauge.GetGeometryRef()
    gauge_x, gauge_y = geom.GetX(), geom.GetY()
    out = grass.read_command('r.what', input=basins, east_north=(gauge_x, gauge_y), stdout = DEVNULL, stderr = DEVNULL)
    cat_grass = out.split('|')[3].strip('\n')
    rules_ws.write(cat_grass.strip() + ':' + cat_grass.strip() + ':' + str(int(gauge.GetField(col_name))) + '\n')
rules_ws.close()
basins_rcl='basins_rcl'
cmd = 'cat '+recl_watersheds_file+' | r.recode --o input='+basins+' output='+basins_rcl
p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) 
p.communicate()
# Save the rules to directory_out
cmd = 'mv '+recl_watersheds_file+' '+directory_out
subprocess.call(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
grass.run_command('r.out.gdal', input=basins_rcl, type='UInt16', output=directory_out + '/' + 'step2_watersheds.tif', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

# Create mask = surface of all watersheds
print('---------- HRU-delin Step 2 : Creating the mask')
grass.run_command('r.null', map=basins_rcl, setnull=0, stdout = DEVNULL, stderr = DEVNULL)
mask_tmp = 'mask_tmp'
cmd = 'echo "0 thru 10000000 = 1" | r.reclass --o input='+basins_rcl+' output='+mask_tmp
p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
p.communicate()
grass.run_command('r.null', map=mask_tmp, null=0, setnull=1, stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.out.gdal', input=mask_tmp, type='UInt16', output=directory_out + '/' + 'step2_mask.tif', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

# Cutting the streams at gauges 
print('---------- HRU-delin Step 2 : Cutting the streams at gauges')
cut_streams_at_gauges()
 
print('---------- HRU-delin Step 2 ended    ---------------------------------------------')
